# -*- Mode: Python; tab-width: 4 -*-
#
# Copyright (C) 2000 Gianluigi Tiesi <sherpya@netfarm.it>
# Copyright (C) 2000 NetFarm S.r.l.  [http://www.netfarm.it]
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# ======================================================================

__version__= '0.99'

from Globals import HTML, DTMLFile      
from Globals import MessageDialog 
from OFS.DTMLDocument import DTMLDocument
from AccessControl import getSecurityManager
from webdav.common import rfc1123_date
from DateTime import DateTime
from sys import platform
from tempfile import mktemp
from os import unlink
from string import find, split
from urllib import quote
import sha

if platform[:3] == "win":
    ### Hack for win32 support
    from os import popen

    class NullHandler:
        def read(self, chars=None):
            return "Win32 Operational Error"
        def close(self):
            pass
    
    class Popen3:
        def __init__(self, cmd, capture=1):
            self.fromchild = popen(cmd, "rb")
            self.tochild = NullHandler()
            self.childerr = NullHandler()

        def wait(self):
            return 0

    swfexe = 'swiftgen.exe'

else:
    from popen2 import Popen3
    swfexe = 'swift-generator'

# self.v_cache_doc
# [0] = Flash File
# [1] = Size of swf file
# [2] = TimeStamp in WebDav Format

FILE=0
SIZE=1
TIMESTAMP=2
DEFAULT_CACHE=1024

html_flash = """<!-- Generated by ZSwGenerator -->
<object classid="clsid:D27CDB6E-AE6D-11cf-96B8-444553540000"
codebase="http://active.macromedia.com/flash2/cabs/swflash.cab#version=4,0,0,0" id="%s" width="%s" height="%s">
<param name="movie" value="%s">
<param name="quality" value="%s">
<param name="bgcolor" value="%s">
<param name="loop" value="%s">
<param name="play" value="%s">
<param name="menu" value="%s">
<embed src="%s" quality="%s" bgcolor="%s" width="%s" height="%s"
loop="%s" play="%s" menu="%s"
type="application/x-shockwave-flash"
pluginspage="http://www.macromedia.com/shockwave/download/index.cgi?P1_Prod_Version=ShockwaveFlash">
</embed>
</object>
<!--            End            -->"""

default = {
'objectid': 'zswgen',
'quality': 'high',
'bgcolor': '#ffffff',
'width': '100%',
'height': '100%',
'loop': 'true',
'play': 'true',
'menu': 'true'
}


manage_addSwGeneratorForm = DTMLFile('dtml/zswgeneratorAddForm', globals())

def manage_addZSwGenerator(self, id, title='ZSwGenerator object',
                  file='', REQUEST=None, submit=None):
    """Add a ZSwGenerator Object """

    if type(file) is not type(''): file=file.read()
    if not file: raise "ZSwError", "Template File is required"
    ob=ZSwGenerator(id, title, file)
    ob.title=title
    id=self._setObject(id, ob)
    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''

class ZSwGenerator(DTMLDocument):
    """ZSwGenerator object."""    
        
    meta_type = 'ZSwGenerator' 

    props=[]
    prop_set=default.items()
    prop_set.sort()
    for key,item in prop_set:
        if type(item)==type(1):
             _tp='int'
        elif type(item)==type(()):
            _tp='tokens'
        else:
            _tp='string'
        props.append({'id':key, 'type': _tp})

    _properties=tuple(props)

        
    def embed(self, **kw):
        fl = {}
        fl['quality'] = self.quality
        fl['bgcolor'] = self.bgcolor
        fl['width'] = self.width
        fl['height'] = self.height
        fl['loop'] = self.loop
        fl['play'] = self.play
        fl['menu'] = self.menu
        fl['objectid'] = self.objectid
        
        num_args = len(kw)
        query = ''
        if num_args:
            for key in kw.keys():
                if key[0] == '_':
                    fl[key[1:]] = kw[key]
                    del kw[key]
                else:
                    param = "%s=%s" % (quote(key), quote(kw[key]))
                    if query == '':
                        query = query + "?" + param
                    else:
                        query = query + "&" + param

        fl['movie'] = self.absolute_url() + query       
        embedded = html_flash % (fl['objectid'],
                                 fl['width'],
                                 fl['height'],
                                 fl['movie'],
                                 fl['quality'],
                                 fl['bgcolor'],
                                 fl['loop'],
                                 fl['play'],
                                 fl['menu'],
                                 fl['movie'],
                                 fl['quality'],
                                 fl['bgcolor'],
                                 fl['width'],
                                 fl['height'],
                                 fl['loop'],
                                 fl['play'],
                                 fl['menu'])
        return embedded
    
    def _clean_sws(self, sws):
        heading = "% Source SwS File"

        # Prune OUTPUT
        src_string = '%OUTPUT -cgi \"-\"'
        src_position = find(sws, src_string)
        sws = sws[(src_position + len(src_string)):]
        
        # Comment for defines

        # Line division
        sws_lines = split(sws, "\n")
        new_sws = heading
        line_prefix = ""
        multiline = 0

        for i in range(len(sws_lines)):
            # If we are in a comment block
            # Check for block end
            if multiline:
                if find(sws_lines[i], "}") != -1:
                    multiline = 0
                else:
                    line_prefix = "% "

            else:
                # If a line begin for DEFINE
                # We comment it out
                if sws_lines[i][:6] == "DEFINE":
                    line_prefix = "% "
                    # If { is in, a block starts
                    if find(sws_lines[i], "{") != -1:
                        multiline = 1
                else:
                    line_prefix = ""
            
            new_sws = new_sws + line_prefix + sws_lines[i] + "\n"

        return new_sws

         
    def _get_sws(self, data):
        sws = ''
        template_file = mktemp() + ".swt"
        try:
            unlink(template_file)
        except: pass
        templ_fd = open(template_file, "wb")
        templ_fd.write(data)
        templ_fd.close()
        cmd_line = "%s -d %s" % (swfexe, template_file)
        mypipe = Popen3(cmd_line, 1)
        mypipe.tochild.close()
        sws = mypipe.fromchild.read()
        mypipe.fromchild.close()
        err = mypipe.childerr.read()
        mypipe.childerr.close()
        res = mypipe.wait()
        del mypipe
        try:
            unlink(template_file)
        except: pass

        if sws == '':
            return None, err
        else:
            sws = self._clean_sws(sws)
            return sws, err

    def _make_swf(self, sws):
        swt = ''
        template_file = mktemp() + ".swt"
        sws_file = mktemp()
        input_str = 'INPUT "%s"\n' % template_file
        output_str = 'OUTPUT "-"'
        sws = input_str + output_str + sws
        try:
            unlink(template_file)
            unlink(sws_file)
        except: pass
        templ_fd = open(template_file, "wb")
        templ_fd.write(self.swt_file)
        templ_fd.close()
        del templ_fd
        sws_fd = open(sws_file, "wb")
        sws_fd.write(sws)
        sws_fd.close()
        del sws_fd
        cmd_line = "%s %s" % (swfexe, sws_file)
        mypipe = Popen3(cmd_line, 1)
        mypipe.tochild.close()
        swt = mypipe.fromchild.read()
        mypipe.fromchild.close()
        err = mypipe.childerr.read()
        mypipe.childerr.close()
        res = mypipe.wait()
        del mypipe
        try:
            unlink(template_file)
            unlink(sws_file)
        except: pass
        
        if swt == '':
            return None, err
        else:        
            return swt, err

    def get_cache_size(self):
        # Cache Size in Kb
        if not hasattr(self,'_v_cache_size'):
            self._v_cache_size = 0
        return int(self._v_cache_size / 1024)
   
    def clean_cache(self):
        # Clean Cache
        self._v_cache_doc = {}
        self._v_cache_size = 0
        self._v_cache_list = []

    def change_cache(self, new_max):
        self.max_cache = new_max
            
    def _to_swf(self, doc_src, qs):
        # First check if there were any modifications

        if not hasattr(self,'_v_cache_doc'):
            self._v_cache_doc = {}
            
        if not hasattr(self,'_v_cache_size'):
            self._v_cache_size = 0

        if not hasattr(self,'_v_cache_list'):
            self._v_cache_list = []
            
        # ID --> query string + document source
        shaobj = sha.new()
        shaobj.update(doc_src)
        shaobj.update(qs)
        qs_index = shaobj.hexdigest()
                        
        if self._v_cache_doc.has_key(qs_index):
            # Raising up last doc request
            self._v_cache_list.remove(qs_index)
            self._v_cache_list.insert(0, qs_index)
            return self._v_cache_doc[qs_index], 'ok'
        
        swf_doc, err = self._make_swf(doc_src)

        if swf_doc:
            # Pseudo-Casting
            self._v_cache_doc[qs_index] = [ 0, 0, '']        
            self._v_cache_doc[qs_index][FILE] = swf_doc
            self._v_cache_doc[qs_index][SIZE] = len(swf_doc)
            self._v_cache_doc[qs_index][TIMESTAMP] = rfc1123_date(DateTime())
            self._v_cache_size = self._v_cache_size + self._v_cache_doc[qs_index][SIZE]
            self._v_cache_list.insert(0, qs_index)
            # Cleaning Up Cache...
            
            while self._v_cache_size > (self.max_cache * 1024):
                to_delete = self._v_cache_list.pop()
                self._v_cache_size = self._v_cache_size - self._v_cache_doc[to_delete][SIZE]
                del self._v_cache_doc[to_delete]
            return self._v_cache_doc[qs_index], 'ok'
        else:
            return None, err

    def __init__(self, id, title, file): 
        """Init of ZSwGenerator"""
        sws, err = self._get_sws(file)
        if not sws:
            raise "ZSwError", ("Error Parsing Template\n<pre>%s</pre>" % err)
        self.swt_file = file
        self.max_cache = DEFAULT_CACHE
        file = sws
        self.sws_source = file
        guessed = 1
        match = None
		## Try to guess width and height
        try:
            match = self.swt_file[find(self.swt_file, "width") + 7:]
            end_pos = find(match, '\"')
            guessed_width = int(match[:end_pos])
            match = match[find(match, "height") + 8:]
            end_pos = find(match, '\"')
            guessed_height = int(match[:end_pos])
        except:
            guessed = None

        del match
            
        DTMLDocument.__init__.im_func(self, file, __name__=id)

        _st={}
        _st.update(default)

        for key,item in _st.items():
            self._updateProperty(key,item)

        self._updateProperty('objectid', id)
        if guessed:
            self._updateProperty('width', guessed_width)
            self._updateProperty('height', guessed_height)

    def get_size(self):
        return len(self.swt_file)
    
    manage_about = DTMLFile('dtml/README', globals())
    manage_cache = DTMLFile('dtml/zswgeneratorCacheForm', globals())
    manage_uploadswtForm = DTMLFile('dtml/zswgeneratorUploadForm', globals())
    manage_main = DTMLFile('dtml/zswgeneratorEdit', globals())

    manage_options=({'label':'Edit', 'action':'manage_main'},
                    {'label':'Upload', 'action':'manage_uploadswtForm'},
                    {'label':'Properties', 'action':'manage_propertiesForm'},
                    {'label':'View', 'action':''},
                    {'label':'Proxy', 'action':'manage_proxyForm'},
                    {'label':'Security', 'action':'manage_access'},
                    {'label':'Cache', 'action':'manage_cache'},
                    {'label':'About', 'action':'manage_about'},
                   )


    __ac_permissions__=( 
    ('Change ZSwGenerator', ('manage_edit', 'manage_upload', 'PUT')),
    ('View management screens', ('manage_tabs', 'manage_main','manage_about','manage_elementHelp')),
    )


    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render the document given a client object, REQUEST mapping,
        Response, and key word arguments."""

        if REQUEST.environ.has_key('QUERY_STRING'):
            qs = REQUEST.environ['QUERY_STRING']
        else:
            qs = ''
                

        kw['document_id'] = self.getId()
        kw['document_title']=self.title
        
        if hasattr(self, 'aq_explicit'):
            bself=self.aq_explicit
        else:
            bself=self

        security=getSecurityManager()
        security.addContext(self)

        try:
            if client is None:
                # Called as subtemplate, so don't need error propigation!
                r=apply(HTML.__call__, (self, bself, REQUEST), kw)
                if RESPONSE is None:
                    result = r
                else:
                    result = decapitate(r, RESPONSE)
                return self._to_swf(result, qs)[0]

            r=apply(HTML.__call__, (self, (client, bself), REQUEST), kw)
            if (type(r) is not type('')) or (RESPONSE is None):
                return self._to_swf(r, qs)[0]

            flash_file, err = self._to_swf(r, qs)
            if flash_file:
                RESPONSE.setHeader('Content-Type','application/x-shockwave-flash')
                RESPONSE.setHeader('Last-Modified', flash_file[2])
                return flash_file[0]
            else:
                raise "ZSwError", ("Problems executing external program\n<pre>%s</pre>" % err)

        finally: security.removeContext(self)

    
    def manage_edit(self,data,title,SUBMIT='Change',
                    dtpref_cols='50',
                    dtpref_rows='20',REQUEST=None):
        """From Zope standard distribution ...
        
        Replaces a Documents contents with Data, Title with Title.

        The SUBMIT parameter is also used to change the size of the editing
        area on the default Document edit screen.  If the value is "Smaller",
        the rows and columns decrease by 5.  If the value is "Bigger", the
        rows and columns increase by 5.  If any other or no value is supplied,
        the data gets checked for DTML errors and is saved.
        """
        
        self._validateProxy(REQUEST)
        if self._size_changes.has_key(SUBMIT):
            return self._er(data,title,SUBMIT,dtpref_cols,dtpref_rows,REQUEST)

        self.title=title

        if SUBMIT == "Reset Script":
            data = self.sws_source
        else:
            if type(data) is not type(''): data=data.read()
        self.munge(data)
        if REQUEST:
            message="Content changed."
            return self.manage_main(self,REQUEST,manage_tabs_message=message)

        
    def manage_uploadswt(self, file='', REQUEST=None):
        """Replace the contents of flash object """
        
        self._validateProxy(REQUEST)
        if type(file) is not type(''): file=file.read()
        sws = None
        err = 'Empty file'
        if len(file) > 0:
            sws, err = self._get_sws(file)

        if sws:
            self.swt_file = file
            self.sws_source = sws
            file = sws
            self.munge(file)

            if REQUEST: return MessageDialog(
                title  ='Success!',
                message='Your changes have been saved',
                action ='manage_main')
        else:
            if REQUEST: return MessageDialog(
                title  ='Error!',
                message='Error Parsing Template\n<pre>%s</pre>' % err,
                action ='manage_main')


